# frozen_string_literal: true

require_relative 'immutable_object'
require_relative 'model_schema'

module Upgrow
  # Base class for Models. As an Immutable Object, it sets a default schema with
  # the minimal attribute ID, as well as requires all attributes to be set upon
  # initialization.
  class BasicModel < ImmutableObject
    class AssociationNotLoadedError < StandardError; end

    class << self
      # Defines an association in the Model Schema.
      #
      # @param name [Symbol] the name of the association.
      def belongs_to(name)
        schema.association(name)
      end

      # Defines an association in the Model Schema.
      #
      # @param name [Symbol] the name of the association.
      def has_many(name)
        schema.association(name)
      end
    end

    self.schema = ModelSchema.new

    attribute :id

    attr_reader :associations

    # Initializes a new Model with the given member values.
    #
    # @param args [Hash<Symbol, Object>] the list of values for each attribute
    #   and association.
    #
    # @raise [KeyError] if an attribute is missing in the list of arguments.
    def initialize(**args)
      @associations = self.class.schema.association_names.to_h do |name|
        [name, args[name]]
      end.freeze

      attributes = self.class.schema.attribute_names.to_h do |name|
        [name, args.fetch(name)]
      end

      super(**attributes)
    end

    # Returns a Hash representation of Model along with all loaded associations.
    #
    # @return [Hash<Symbol, Object>] the list of attribute names
    #   and associations, if loaded.
    def to_h
      associations_hash = associations.compact.transform_values do |models|
        models.respond_to?(:map) ? models.map(&:to_h) : models.to_h
      end
      attributes.merge(associations_hash)
    end

    private

    def method_missing(name, *args, &block)
      return super unless associations.include?(name)

      associations[name] || raise(
        AssociationNotLoadedError,
        "Association #{name} not loaded for #{self.class.name}."
      )
    end

    def respond_to_missing?(name, _include_private = false)
      associations.include?(name) || super
    end
  end
end
